# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import os
import pytest
import logging
import uuid
import warnings
import datetime
from functools import partial
from logging.handlers import RotatingFileHandler
from azure.core.settings import settings

from azure.mgmt.eventhub import EventHubManagementClient
from azure.eventhub import EventHubProducerClient, TransportType
from azure.eventhub._pyamqp import ReceiveClient
from azure.eventhub._pyamqp.authentication import SASTokenAuth
from azure.eventhub.extensions.checkpointstoreblob import BlobCheckpointStore
from azure.eventhub.extensions.checkpointstoreblobaio import BlobCheckpointStore as BlobCheckpointStoreAsync

try:
    import uamqp

    uamqp_transport_params = [True, False]
    uamqp_transport_ids = ["uamqp", "pyamqp"]
except (ModuleNotFoundError, ImportError):
    uamqp_transport_params = [False]
    uamqp_transport_ids = ["pyamqp"]

socket_transports = [TransportType.Amqp, TransportType.AmqpOverWebsocket]
socket_transport_ids = ["amqp", "ws"]

from devtools_testutils import get_region_override, get_credential as get_devtools_credential
from tracing_common import FakeSpan

collect_ignore = []
PARTITION_COUNT = 2
CONN_STR = "Endpoint=sb://{}/;SharedAccessKeyName={};SharedAccessKey={};EntityPath={}"
RES_GROUP_PREFIX = "eh-res-group"
NAMESPACE_PREFIX = "eh-ns"
EVENTHUB_PREFIX = "eh"
EVENTHUB_DEFAULT_AUTH_RULE_NAME = "RootManageSharedAccessKey"
LOCATION = get_region_override("westus")


def pytest_addoption(parser):
    parser.addoption("--sleep", action="store", default="True", help="sleep on reconnect test: True or False")


@pytest.fixture
def sleep(request):
    sleep = request.config.getoption("--sleep")
    return sleep.lower() in ("true", "yes", "1", "y")


@pytest.fixture(scope="session", params=uamqp_transport_params, ids=uamqp_transport_ids)
def uamqp_transport(request):
    return request.param


@pytest.fixture(scope="session", params=socket_transports, ids=socket_transport_ids)
def socket_transport(request):
    return request.param


@pytest.fixture(scope="session")
def storage_url():
    try:
        account_name = os.environ["AZURE_STORAGE_ACCOUNT"]
        return f"https://{account_name}.blob.core.windows.net"
    except KeyError:
        warnings.warn(UserWarning("AZURE_STORAGE_ACCOUNT undefined"))
        pytest.skip("AZURE_STORAGE_ACCOUNT undefined")
        return


@pytest.fixture()
def checkpoint_store(storage_url):
    checkpoint_store = BlobCheckpointStore(
        storage_url, "blobcontainer" + str(uuid.uuid4()), credential=get_devtools_credential()
    )
    return checkpoint_store


@pytest.fixture()
def checkpoint_store_aio(storage_url):
    checkpoint_store = BlobCheckpointStoreAsync(
        storage_url, "blobcontainer" + str(uuid.uuid4()), credential=get_devtools_credential(is_async=True)
    )
    return checkpoint_store


def get_logger(filename, level=logging.INFO):
    azure_logger = logging.getLogger("azure.eventhub")
    azure_logger.setLevel(level)
    uamqp_logger = logging.getLogger("uamqp")
    uamqp_logger.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(formatter)
    if not azure_logger.handlers:
        azure_logger.addHandler(console_handler)
    if not uamqp_logger.handlers:
        uamqp_logger.addHandler(console_handler)

    if filename:
        file_handler = RotatingFileHandler(filename, maxBytes=5 * 1024 * 1024, backupCount=2)
        file_handler.setFormatter(formatter)
        azure_logger.addHandler(file_handler)

    return azure_logger


log = get_logger(None, logging.DEBUG)


@pytest.fixture(scope="session")
def timeout_factor(uamqp_transport):
    if uamqp_transport:
        return 1000
    else:
        return 1


@pytest.fixture(scope="session")
def fake_span():
    return FakeSpan


@pytest.fixture()
def enable_tracing(fake_span):
    settings.tracing_implementation.set_value(fake_span)
    yield fake_span
    # MUST RESET TRACING IMPLEMENTATION TO NONE, ELSE ALL OTHER TESTS ADD TRACING
    settings.tracing_implementation.set_value(None)


@pytest.fixture(scope="session")
def get_credential():
    return get_devtools_credential


@pytest.fixture(scope="session")
def get_credential_async():
    return partial(get_devtools_credential, is_async=True)


@pytest.fixture(scope="session")
def resource_group():
    try:
        resource_group_name = os.environ["EVENTHUB_RESOURCE_GROUP"]
    except KeyError:
        warnings.warn(UserWarning("EVENTHUB_RESOURCE_GROUP undefined - skipping test"))
        pytest.skip("EVENTHUB_RESOURCE_GROUP defined")
    return resource_group_name


@pytest.fixture(scope="session")
def eventhub_namespace(resource_group):
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")
        return
    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes
    )
    try:
        namespace_name = os.environ["EVENT_HUB_NAMESPACE"]
    except KeyError:
        warnings.warn(UserWarning("EVENT_HUB_NAMESPACE undefined - skipping test"))
        pytest.skip("EVENT_HUB_NAMESPACE defined")

    key = resource_client.namespaces.list_keys(resource_group, namespace_name, EVENTHUB_DEFAULT_AUTH_RULE_NAME)
    connection_string = key.primary_connection_string
    key_name = key.key_name
    primary_key = key.primary_key
    return namespace_name, connection_string, key_name, primary_key


@pytest.fixture()
def live_eventhub(resource_group, eventhub_namespace):  # pylint: disable=redefined-outer-name
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        warnings.warn(UserWarning("AZURE_SUBSCRIPTION_ID undefined - skipping test"))
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")

    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes
    )
    eventhub_name = EVENTHUB_PREFIX + str(uuid.uuid4())
    eventhub_ns_name, connection_string, key_name, primary_key = eventhub_namespace
    eventhub_endpoint_suffix = os.environ.get("EVENT_HUB_ENDPOINT_SUFFIX", ".servicebus.windows.net")
    try:
        eventhub = resource_client.event_hubs.create_or_update(
            resource_group, eventhub_ns_name, eventhub_name, {"partition_count": PARTITION_COUNT}
        )
        live_eventhub_config = {
            "resource_group": resource_group,
            "hostname": "{}{}".format(eventhub_ns_name, eventhub_endpoint_suffix),
            "key_name": key_name,
            "access_key": primary_key,
            "namespace": eventhub_ns_name,
            "event_hub": eventhub.name,
            "consumer_group": "$Default",
            "partition": "0",
            "connection_str": connection_string + ";EntityPath=" + eventhub.name,
        }
        yield live_eventhub_config
    finally:
        try:
            resource_client.event_hubs.delete(resource_group, eventhub_ns_name, eventhub_name)
        except:
            warnings.warn(UserWarning("eventhub teardown failed"))


@pytest.fixture()
def resource_mgmt_client():
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        warnings.warn(UserWarning("AZURE_SUBSCRIPTION_ID undefined - skipping test"))
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")

    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes
    )
    yield resource_client


@pytest.fixture()
def connection_str(live_eventhub):
    return CONN_STR.format(
        live_eventhub["hostname"], live_eventhub["key_name"], live_eventhub["access_key"], live_eventhub["event_hub"]
    )


@pytest.fixture()
def invalid_hostname(live_eventhub):
    return CONN_STR.format(
        "invalid123.servicebus.windows.net",
        live_eventhub["key_name"],
        live_eventhub["access_key"],
        live_eventhub["event_hub"],
    )


@pytest.fixture()
def invalid_key(live_eventhub):
    return CONN_STR.format(live_eventhub["hostname"], live_eventhub["key_name"], "invalid", live_eventhub["event_hub"])


@pytest.fixture()
def invalid_policy(live_eventhub):
    return CONN_STR.format(
        live_eventhub["hostname"], "invalid", live_eventhub["access_key"], live_eventhub["event_hub"]
    )


@pytest.fixture()
def connstr_receivers(live_eventhub, uamqp_transport):
    connection_str = live_eventhub["connection_str"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"], live_eventhub["event_hub"], live_eventhub["consumer_group"], p
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(source, auth=sas_auth, debug=False, timeout=0, prefetch=500)
        else:
            sas_auth = SASTokenAuth(uri, uri, live_eventhub["key_name"], live_eventhub["access_key"])
            receiver = ReceiveClient(
                live_eventhub["hostname"], source, auth=sas_auth, network_trace=False, timeout=0, link_credit=500
            )
        receiver.open()
        receivers.append(receiver)
    yield connection_str, receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credentials(live_eventhub):
    return live_eventhub["hostname"], live_eventhub["event_hub"], get_devtools_credential


@pytest.fixture()
def auth_credentials_async(live_eventhub):
    return live_eventhub["hostname"], live_eventhub["event_hub"], partial(get_devtools_credential, is_async=True)


@pytest.fixture()
def auth_credential_receivers(live_eventhub, uamqp_transport):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"], live_eventhub["event_hub"], live_eventhub["consumer_group"], p
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(source, auth=sas_auth, debug=False, timeout=0, prefetch=500)
        else:
            # TODO: TokenAuth should be fine?
            sas_auth = SASTokenAuth(uri, uri, live_eventhub["key_name"], live_eventhub["access_key"])
            receiver = ReceiveClient(
                live_eventhub["hostname"], source, auth=sas_auth, network_trace=False, timeout=30, link_credit=500
            )
        receiver.open()
        receivers.append(receiver)
    yield fully_qualified_namespace, eventhub_name, get_devtools_credential, receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credential_receivers_async(live_eventhub, uamqp_transport):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"], live_eventhub["event_hub"], live_eventhub["consumer_group"], p
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(source, auth=sas_auth, debug=False, timeout=0, prefetch=500)
        else:
            # TODO: TokenAuth should be fine?
            sas_auth = SASTokenAuth(uri, uri, live_eventhub["key_name"], live_eventhub["access_key"])
            receiver = ReceiveClient(
                live_eventhub["hostname"], source, auth=sas_auth, network_trace=False, timeout=30, link_credit=500
            )
        receiver.open()
        receivers.append(receiver)
    yield fully_qualified_namespace, eventhub_name, partial(get_devtools_credential, is_async=True), receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credential_senders(live_eventhub, uamqp_transport):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=get_devtools_credential(),
        uamqp_transport=uamqp_transport,
    )
    partitions = client.get_partition_ids()

    senders = []
    for p in partitions:
        sender = client._create_producer(partition_id=p)
        senders.append(sender)
    yield fully_qualified_namespace, eventhub_name, get_devtools_credential, senders
    for s in senders:
        s.close()
    client.close()


@pytest.fixture()
def auth_credential_senders_async(live_eventhub, uamqp_transport):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=get_devtools_credential(),
        uamqp_transport=uamqp_transport,
    )
    partitions = client.get_partition_ids()

    senders = []
    for p in partitions:
        sender = client._create_producer(partition_id=p)
        senders.append(sender)
    yield fully_qualified_namespace, eventhub_name, partial(get_devtools_credential, is_async=True), senders
    for s in senders:
        s.close()
    client.close()


# Note: This is duplicated between here and the basic conftest, so that it does not throw warnings if you're
# running locally to this SDK. (Everything works properly, pytest just makes a bit of noise.)
def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line("markers", "liveTest: mark test to be a live test only")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

class WebSocketException(Exception):
    """
    Base exception for all WebSocket related errors.
    """

class WebSocketPayloadError(WebSocketException):
    """
    Raised when there is an error in the WebSocket payload.
    """

class WebSocketConnectionError(WebSocketException):
    """
    Raised when there is an eror while establishing a connection.
    """

class WebSocketConnectionClosed(WebSocketConnectionError):
    """
    Raised when the connection is closed.
    """

class WebSocketProtocolError(WebSocketException):
    """
    Raised when the WebSocket protocol is violated.
    """



from unittest.mock import Mock
from azure.eventhub._pyamqp.error import AMQPLinkError
from azure.eventhub._pyamqp.link import Link
from azure.eventhub._pyamqp.receiver import ReceiverLink
from azure.eventhub._pyamqp.constants import LinkState
import pytest


@pytest.mark.parametrize(
    "start_state,expected_state",
    [
        (LinkState.ATTACHED, LinkState.DETACH_SENT),
        (LinkState.ATTACH_SENT, LinkState.DETACHED),
        (LinkState.ATTACH_RCVD, LinkState.DETACHED),
    ],
    ids=["link attached", "link attach sent", "link attach rcvd"],
)
def test_link_should_detach(start_state, expected_state):
    session = Mock()
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    link._set_state(start_state)
    link._outgoing_detach = Mock(return_value=None)
    link.detach()

    assert link.state == expected_state


@pytest.mark.parametrize(
    "state",
    [LinkState.DETACHED, LinkState.DETACH_SENT, LinkState.ERROR],
    ids=["link detached", "link detach sent", "link error"],
)
def test_link_should_not_detach(state):
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    link._set_state(state)
    link._outgoing_detach = Mock(return_value=None)
    link.detach()
    link._outgoing_detach.assert_not_called()


def test_receive_transfer_frame_multiple():
    session = None
    link = ReceiverLink(
        session,
        3,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
        on_transfer=Mock(),
    )

    link.current_link_credit = 2  # Set the link credit to 2

    # frame: handle, delivery_id, delivery_tag, message_format, settled, more, rcv_settle_mode, state, resume, aborted, bathable, payload
    transfer_frame_one = [3, 0, b"/blah", 0, True, True, None, None, None, None, False, ""]
    transfer_frame_two = [3, None, b"/blah", 0, True, False, None, None, None, None, False, ""]

    link._incoming_transfer(transfer_frame_one)
    assert link.current_link_credit == 2
    link._incoming_transfer(transfer_frame_two)
    assert link.current_link_credit == 1


def test_receive_transfer_continuation_frame():
    session = None
    link = ReceiverLink(
        session,
        3,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
        on_transfer=Mock(),
    )

    link.current_link_credit = 3  # Set the link credit to 2

    # frame: handle, delivery_id, delivery_tag, message_format, settled, more, rcv_settle_mode, state, resume, aborted, batchable, payload
    transfer_frame_one = [3, 0, b"/blah", 0, True, False, None, None, None, None, False, ""]
    transfer_frame_two = [3, 1, b"/blah", 0, True, True, None, None, None, None, False, ""]
    transfer_frame_three = [3, None, b"/blah", 0, True, False, None, None, None, None, False, ""]

    link._incoming_transfer(transfer_frame_one)
    assert link.current_link_credit == 2
    assert link.delivery_count == 1
    link._incoming_transfer(transfer_frame_two)
    assert link.current_link_credit == 2
    assert link.delivery_count == 1
    link._incoming_transfer(transfer_frame_three)
    assert link.current_link_credit == 1
    assert link.delivery_count == 2


def test_receive_transfer_and_flow():
    def mock_outgoing():
        pass

    session = None
    link = ReceiverLink(
        session,
        3,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
        on_transfer=Mock(),
    )

    link._outgoing_flow = mock_outgoing
    link.total_link_credit = 0  # Set the total link credit to 0 to start, no credit on the wire

    link.flow(link_credit=100)  # Send a flow frame with desired link credit of 100

    # frame: handle, delivery_id, delivery_tag, message_format, settled, more, rcv_settle_mode, state, resume, aborted, batchable, payload
    transfer_frame_one = [3, 0, b"/blah", 0, True, False, None, None, None, None, False, ""]
    transfer_frame_two = [3, 1, b"/blah", 0, True, False, None, None, None, None, False, ""]
    transfer_frame_three = [3, 2, b"/blah", 0, True, False, None, None, None, None, False, ""]

    link._incoming_transfer(transfer_frame_one)
    assert link.current_link_credit == 99
    assert link.total_link_credit == 99

    # Only received 1 transfer frame per receive call, we set desired link credit again
    # this will send a flow of 1
    link.flow(link_credit=100)
    assert link.current_link_credit == 1
    assert link.total_link_credit == 100

    link._incoming_transfer(transfer_frame_two)
    assert link.current_link_credit == 0
    assert link.total_link_credit == 99
    link._incoming_transfer(transfer_frame_three)
    assert link.current_link_credit == -1
    assert link.total_link_credit == 98

@pytest.mark.parametrize(
    "frame",
    [
        [2, True, [b'amqp:link:detach-forced', b"The link is force detached. Code: publisher(link3006875). Details: AmqpMessagePublisher.IdleTimerExpired: Idle timeout: 00:10:00.", None]],
        [2, True, [b'amqp:link:detach-forced', None, b'something random']],
        [2, True, [b'amqp:link:detach-forced', None, None]],
        [2, True, [b'amqp:link:detach-forced']],
        
    ],
    ids=["description and info", "info only", "description only", "no info or description"],
)
def test_detach_with_error(frame):
    '''
      A detach can optionally include an description and info field.
      https://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-transport-v1.0-os.html#type-error
    '''
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    link._set_state(LinkState.DETACH_RCVD)
    link._incoming_detach(frame)

    with pytest.raises(AMQPLinkError) as ae:
        link.get_state()
        assert ae.description == frame[2][1]
        assert ae.info == frame[2][2]






# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from urllib.parse import ParseResult, urlparse
from dataclasses import dataclass


@dataclass
class WebSocketURL:
    scheme: str
    hostname: str
    port: int
    path: str
    query: str
    fragment: str
    is_secure: bool

def parse_url(url: str) -> WebSocketURL:
    parsed_url: ParseResult = urlparse(url)
    if parsed_url.scheme not in ('ws', 'wss'):
        raise ValueError('Scheme must be ws or wss')
    if not parsed_url.hostname:
        raise ValueError('Provided hostname is invalid')
    if not parsed_url.port:
        port = 80 if parsed_url.scheme == 'ws' else 443
    if not parsed_url.path:
        path = '/'
    else:
        path = parsed_url.path

    return WebSocketURL(
        scheme=parsed_url.scheme,
        hostname=parsed_url.hostname,
        port=port if not parsed_url.port else parsed_url.port,
        path=path,
        query=parsed_url.query,
        fragment=parsed_url.fragment,
        is_secure=parsed_url.scheme == 'wss'
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from dataclasses import dataclass
import io
import os
import struct
import enum

from ._exceptions import WebSocketPayloadError, WebSocketProtocolError
from ._utils import mask_payload


class Opcode(enum.IntEnum):
    CONTINUATION = 0x00
    TEXT = 0x01
    BINARY = 0x02
    CLOSE = 0x08
    PING = 0x09
    PONG = 0x0A

FRAMES = {
    Opcode.CONTINUATION,
    Opcode.TEXT,
    Opcode.BINARY,
    Opcode.CLOSE,
    Opcode.PING,
    Opcode.PONG
}

CONTROL_FRAMES = {
    Opcode.CLOSE,
    Opcode.PING,
    Opcode.PONG,
}

DATA_FRAMES = {
    Opcode.TEXT,
    Opcode.BINARY,
    Opcode.CONTINUATION,
}


class CloseReason(enum.IntEnum):
    '''
    The status codes for the close frame as per RFC 6455.
    https://datatracker.ietf.org/doc/html/rfc6455#section-7.4.1
   '''
    NORMAL = 1000
    GOING_AWAY = 1001
    PROTOCOL_ERROR = 1002
    UNSUPPORTED_DATA = 1003
    NO_STATUS_RCVD = 1005
    ABNORMAL_CLOSURE = 1006
    INVALID_PAYLOAD = 1007
    POLICY_VIOLATION = 1008
    MESSAGE_TOO_BIG = 1009
    MANDATORY_EXT = 1010
    INTERNAL_ERROR = 1011
    SERVICE_RESTART = 1012
    TRY_AGAIN_LATER = 1013
    BAD_GATEWAY = 1014
    TLS_HANDSHAKE = 1015

ALLOWED_CLOSED_REASONS = {
    CloseReason.NORMAL,
    CloseReason.GOING_AWAY,
    CloseReason.PROTOCOL_ERROR,
    CloseReason.UNSUPPORTED_DATA,
    CloseReason.INVALID_PAYLOAD,
    CloseReason.POLICY_VIOLATION,
    CloseReason.MESSAGE_TOO_BIG,
    CloseReason.MANDATORY_EXT,
    CloseReason.INTERNAL_ERROR,
    CloseReason.SERVICE_RESTART,
    CloseReason.TRY_AGAIN_LATER,
    CloseReason.BAD_GATEWAY,
}


@dataclass
class Frame:
    """
    Represents a WebSocket frame
    """
    data: bytes
    opcode: int = Opcode.TEXT
    fin: bool = True
    mask: bool = True
    rsv1: int = 0
    rsv2: int = 0
    rsv3: int = 0

    def encode(self) -> bytes:
        output = io.BytesIO()

        header = output.write(struct.pack('!B', # pylint: disable=unused-variable
                             (
                                    self.fin << 7 |  # FIN
                                    self.rsv1 << 6 |  # RSV1
                                    self.rsv2 << 5 |  # RSV2
                                    self.rsv3 << 4 |  # RSV3
                                    self.opcode
                             )))
        mask_bit = 1 << 7
        length = len(self.data)
        payload = bytearray(self.data)

        if length < 126:
            output.write(struct.pack('!B', mask_bit | length))
        elif length < 65536:
            output.write(struct.pack('!BH', mask_bit | 126, length))
        else:
            output.write(struct.pack('!BQ', mask_bit | 127, length))

        if self.mask:
            masking_key = os.urandom(4)
            output.write(masking_key)
            mask_payload(masking_key, payload)
        output.write(payload)

        return output.getvalue()

    def validate(self) -> None:
        """
        Validate the frame according to the RFC 6455
        """
        if self.rsv1 or self.rsv2 or self.rsv3:
            raise WebSocketProtocolError("Reserved bits are set")

        if self.opcode not in FRAMES:
            raise WebSocketProtocolError(f"Invalid opcode {self.opcode}. Expected one of {FRAMES}")

        # Validations for control frames
        # https://datatracker.ietf.org/doc/html/rfc6455#section-5.5

        if self.opcode in CONTROL_FRAMES:
            if not self.fin:
                raise WebSocketProtocolError("Control frames must not be fragmented")

            if len(self.data) > 125:
                raise WebSocketProtocolError(f"Control frames must not be larger than 125 bytes. Got {len(self.data)}")

            if self.opcode == Opcode.CLOSE:
                if not self.data:
                    return

                payload_length = len(self.data)

                # if there is a body, it must be at least 2 bytes long in order to contain the status code
                # https://datatracker.ietf.org/doc/html/rfc6455#section-5.5.1
                if payload_length <= 1:
                    raise WebSocketProtocolError("Close frame with payload length of less than 2 bytes")

                code = struct.unpack('!H', self.data[:2])[0]
                payload = self.data[2:]

                if not (code in ALLOWED_CLOSED_REASONS or 3000 <= code <= 4999):
                    raise WebSocketProtocolError(f"Invalid close code {code}")

                try:
                    payload.decode("utf-8")
                except UnicodeDecodeError as ude:
                    raise WebSocketPayloadError('Invalid UTF-8 payload') from ude

        elif self.opcode == Opcode.TEXT and self.fin:
            try:
                self.data.decode("utf-8")
            except UnicodeDecodeError as ude:
                raise WebSocketPayloadError('Invalid UTF-8 payload') from ude



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import asyncio
import os
import dotenv
import logging
from logging.handlers import RotatingFileHandler
import time

from azure.eventhub.aio import EventHubConsumerClient
from azure.identity.aio import DefaultAzureCredential

logger = logging.getLogger("ASYNC_RECEIVE_PERF_TEST")
logger.setLevel(logging.INFO)
logger.addHandler(RotatingFileHandler("async_receive_perf_test.log"))

dotenv.load_dotenv()
NAMESPACES = [
    os.environ["EVENT_HUB_NAMESPACE_BASIC_NORTHEU"],
    os.environ["EVENT_HUB_NAMESPACE_STANDARD_NORTHEU"],
    os.environ["EVENT_HUB_NAMESPACE_BASIC_WESTUS2"],
    os.environ["EVENT_HUB_NAMESPACE_STANDARD_WESTUS2"],
]
EH_NAME_EVENT_SIZE_PAIR = [
    ("pyamqp_512", 512),
]

PREFETCH_LIST = [300, 3000]
PARTITION_ID = "0"
RUN_DURATION = 30
FIXED_AMOUNT = 100_000


async def receive_fixed_time_interval(
    fully_qualified_namespace,
    eventhub_name,
    credential,
    single_event_size,
    prefetch=300,
    batch_receiving=False,
    description=None,
    run_duration=30,
    partition_id="0",
):
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential,
        consumer_group="$Default",
    )

    last_received_count = [0]
    received_count = [0]
    run_flag = [True]
    all_perf_records = []
    check_interval = 1

    async def on_event(partition_context, event):
        received_count[0] += 1

    async def on_event_batch(partition_context, events):
        received_count[0] += len(events)

    async def monitor():
        while run_flag[0]:
            snap = received_count[0]
            perf = (snap - last_received_count[0]) / check_interval
            last_received_count[0] = snap
            all_perf_records.append(perf)
            await asyncio.sleep(check_interval)

    target = consumer_client.receive_batch if batch_receiving else consumer_client.receive
    kwargs = {
        "partition_id": partition_id,
        "starting_position": "-1",  # "-1" is from the beginning of the partition.
        "prefetch": prefetch,
    }
    if batch_receiving:
        kwargs["max_batch_size"] = prefetch
        kwargs["on_event_batch"] = on_event_batch
    else:
        kwargs["on_event"] = on_event

    recv_future = asyncio.create_task(target(**kwargs))
    monitor_future = asyncio.create_task(monitor())

    await asyncio.sleep(run_duration)
    await consumer_client.close()
    run_flag[0] = False
    await recv_future
    await monitor_future

    valid_perf_records = all_perf_records[10:]  # skip the first 10 records to let the receiving program be stable
    avg_perf = sum(valid_perf_records) / len(valid_perf_records)

    logger.info(
        "EH Namespace: {}.\nMethod: {}, The average performance is {} events/s, throughput: {} bytes/s.\n"
        "Configs are: Single message size: {} bytes, Run duration: {} seconds, Batch: {}.\n"
        "Prefetch: {}".format(
            fully_qualified_namespace,
            description or "receive_fixed_time_interval",
            avg_perf,
            avg_perf * single_event_size,
            single_event_size,
            run_duration,
            batch_receiving,
            prefetch,
        )
    )


async def receive_fixed_amount(
    fully_qualified_namespace,
    eventhub_name,
    credential,
    single_event_size,
    prefetch=300,
    batch_receiving=False,
    description=None,
    partition_id="0",
    run_times=1,
    fixed_amount=100_000,
):
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential,
        consumer_group="$Default",
    )
    perf_records = []
    received_count = [0]

    async def on_event(partition_context, event):
        received_count[0] += 1
        if received_count[0] == fixed_amount:
            await consumer_client.close()

    async def on_event_batch(partition_context, events):
        received_count[0] += len(events)
        if received_count[0] >= fixed_amount:
            await consumer_client.close()

    for i in range(run_times):
        start_time = time.time()
        async with consumer_client:
            if batch_receiving:
                await consumer_client.receive_batch(
                    on_event_batch=on_event_batch,
                    partition_id=partition_id,
                    starting_position="-1",
                    max_batch_size=prefetch,
                    prefetch=prefetch,
                )
            else:
                await consumer_client.receive(
                    on_event=on_event, partition_id=partition_id, starting_position="-1", prefetch=prefetch
                )
        end_time = time.time()
        total_time = end_time - start_time
        speed = fixed_amount / total_time
        perf_records.append(speed)
        received_count[0] = 0
    avg_perf = sum(perf_records) / len(perf_records)

    logger.info(
        "EH Namespace: {}.\nMethod: {}, The average performance is {} events/s, throughput: {} bytes/s.\n"
        "Configs are: Single message size: {} bytes, Total events to receive: {}, Batch: {}.\n"
        "Prefetch: {}".format(
            fully_qualified_namespace,
            description or "receive_fixed_amount",
            avg_perf,
            avg_perf * single_event_size,
            single_event_size,
            fixed_amount,
            batch_receiving,
            prefetch,
        )
    )


if __name__ == "__main__":
    for namespace in NAMESPACES:
        for eh_name, single_event_size in EH_NAME_EVENT_SIZE_PAIR:
            for prefetch in PREFETCH_LIST:
                for batch_receiving in [True, False]:
                    print("------------------- receiving fixed amount -------------------")
                    asyncio.run(
                        receive_fixed_amount(
                            fully_qualified_namespace=namespace,
                            eventhub_name=eh_name,
                            credential=DefaultAzureCredential(),
                            single_event_size=single_event_size,
                            prefetch=prefetch,
                            batch_receiving=batch_receiving,
                            fixed_amount=FIXED_AMOUNT,
                        )
                    )
                    print("------------------- receiving fixed interval -------------------")
                    asyncio.run(
                        receive_fixed_time_interval(
                            fully_qualified_namespace=namespace,
                            eventhub_name=eh_name,
                            credential=DefaultAzureCredential(),
                            single_event_size=single_event_size,
                            prefetch=prefetch,
                            batch_receiving=batch_receiving,
                            run_duration=RUN_DURATION,
                        )
                    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from hashlib import sha1
import os
import base64
from typing import Dict, List, Tuple, Optional
from ._constants import WS_KEY


def build_request_headers(
        resource: str,
        host: str,
        port: int,
        is_secure: bool,
        key: bytes,
        *,
        subprotocols: Optional[List[str]] = None
    ) -> bytes:
    request: List[bytes] = [
        f'GET {resource} HTTP/1.1'.encode(),
    ]
    origin: str = 'https://' if is_secure else 'http://'
    origin = f'{origin}{build_host(host, port)}'

    headers: Dict[bytes, bytes] = {
        b'Host: ': build_host(host, port).encode(),
        b'Connection: ': b'Upgrade',
        b'Upgrade: ': b'websocket',
        b'Origin: ': origin.encode(),
        b'Sec-WebSocket-Key: ': key,
        b'Sec-WebSocket-Version: ': b'13'
    }

    if subprotocols:
        headers[b'Sec-WebSocket-Protocol: '] = ','.join(subprotocols).encode()

    for header, value in headers.items():
        request.append(header + value)

    request.append(b'\r\n')
    request_bytes: bytes = b'\r\n'.join(request)
    return request_bytes

def build_host(host: str, port: int) -> str:
    if ':' in host:
        return f'[{host}]:{port}'

    if port in (80, 443):
        return host

    return f'{host}:{port}'

def build_key() -> bytes:
    random_bytes: bytes = os.urandom(16)
    return base64.b64encode(random_bytes)

def match_key(client_header_key: bytes, server_header_key: bytes) -> bool:
    # the use of sha1 is a websocket protocol requirement and is used for hashing and not cryptography
    # other impls have the same behavior
    # websocket spec: https://datatracker.ietf.org/doc/html/rfc6455
    # https://github.com/dotnet/runtime/blob/45caaf85faa654114f7a3744910df86d8e92882f/src/libraries/System.Net.HttpListener/src/System/Net/WebSockets/HttpWebSocket.cs#L18
    match = base64.b64encode(sha1(client_header_key + WS_KEY).digest()).lower() # nosec

    return match == server_header_key

def parse_response_headers(response_headers: bytes) -> Tuple[bytes, bytes,  Dict[bytes, bytes]]:
    response_line = response_headers.split(b'\r\n')[0]
    _, code, status = response_line.split(b' ', 2)

    headers: Dict[bytes, bytes] = {}

    for header in response_headers.split(b'\r\n')[1:]:
        if not header:
            break
        key, value = header.split(b': ')
        headers[key.lower()] = value.lower()

    return code, status, headers

def parse_proxy_response(response_headers: bytes) -> Tuple[bytes, bytes, bytes]:
    if not response_headers:
        return b'', b'', b''

    response_line = response_headers.split(b'\r\n')[0]

    try:
        version, status, reason = response_line.split(b' ', 2)
    except ValueError:
        try:
            version, status = response_line.split(b' ', 1)
            reason = b''
        except ValueError:
            version = b''

    return version, status, reason



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import functools
from typing import List

@functools.lru_cache
def _xor_table() -> List[bytes]:
    return [bytes(a ^ b for a in range(256)) for b in range(256)]


def mask_payload(masking_key: bytes, data: bytearray) -> None:
    """XOR mask bytes.

    :param masking_key: The masking key.
    :type masking_key: bytes
    :param data: The data to mask.
    :type data: bytearray
    """
    _XOR_TABLE = _xor_table()
    a, b, c, d = (_XOR_TABLE[n] for n in bytearray(masking_key))
    data[::4] = data[::4].translate(a)
    data[1::4] = data[1::4].translate(b)
    data[2::4] = data[2::4].translate(c)
    data[3::4] = data[3::4].translate(d)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events with different options to an Event Hub partition.
"""

import time
import os
from azure.eventhub import EventHubProducerClient, EventData
from azure.eventhub.exceptions import EventHubError
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


# [START send_event_data_batch]
def send_event_data_batch(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch = producer.create_batch()
    event_data_batch.add(EventData("Single message"))
    producer.send_batch(event_data_batch)


# [END send_event_data_batch]


def send_event_data_batch_with_limited_size(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch_with_limited_size = producer.create_batch(max_size_in_bytes=1000)

    while True:
        try:
            event_data_batch_with_limited_size.add(EventData("Message inside EventBatchData"))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break

    producer.send_batch(event_data_batch_with_limited_size)


def send_event_data_batch_with_partition_key(producer):
    # Specifying partition_key.
    event_data_batch_with_partition_key = producer.create_batch(partition_key="pkey")
    event_data_batch_with_partition_key.add(
        EventData("Message will be sent to a partition determined by the partition key")
    )

    producer.send_batch(event_data_batch_with_partition_key)


def send_event_data_batch_with_partition_id(producer):
    # Specifying partition_id.
    event_data_batch_with_partition_id = producer.create_batch(partition_id="0")
    event_data_batch_with_partition_id.add(EventData("Message will be sent to target-id partition"))

    producer.send_batch(event_data_batch_with_partition_id)


def send_event_data_batch_with_properties(producer):
    event_data_batch = producer.create_batch()
    event_data = EventData("Message with properties")
    event_data.properties = {"prop_key": "prop_value"}
    event_data_batch.add(event_data)
    producer.send_batch(event_data_batch)


def send_event_data_list(producer):
    # If you know beforehand that the list of events you have will not exceed the
    # size limits, you can use the `send_batch()` api directly without creating an EventDataBatch

    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.

    event_data_list = [EventData("Event Data {}".format(i)) for i in range(10)]
    try:
        producer.send_batch(event_data_list)
    except ValueError:  # Size exceeds limit. This shouldn't happen if you make sure before hand.
        print("Size of the event data list exceeds the size limit of a single send")
    except EventHubError as eh_err:
        print("Sending error: ", eh_err)


producer = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
)

start_time = time.time()
with producer:
    send_event_data_batch(producer)
    send_event_data_batch_with_limited_size(producer)
    send_event_data_batch_with_partition_key(producer)
    send_event_data_batch_with_partition_id(producer)
    send_event_data_batch_with_properties(producer)
    send_event_data_list(producer)

print("Send messages in {} seconds.".format(time.time() - start_time))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from enum import Enum
from typing import Literal


WS_VERSION: Literal[13] = 13
WS_KEY = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

EOF = b'\r\n\r\n'

class ConnectionStatus(Enum):
    # a client opens a connection and sends a handshake
    CONNECTING = 1

    # a client has successfully completed & validated the handshake response from the server
    OPEN = 2

    # upon either sending or receiving a Close control frame
    CLOSING = 3

    # when the underlying TCP connection is closed
    CLOSED = 4



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Dict, Optional, List, Union, TYPE_CHECKING
from ._protocol import WebSocketProtocol, AsyncWebSocketProtocol
from ._exceptions import WebSocketConnectionError
from ._frame import Opcode, CloseReason, Frame

if TYPE_CHECKING:
    from ssl import SSLContext


class WebSocket:
    """
    A synchronous WebSocket client providing a simple interface for sending and receiving data.

    :param str url: The WebSocket server URL, starting with ws:// or wss:// in the format 
    ws://host:port/path or wss://host:port/path.
    :keyword dict[bytes, bytes] headers: HTTP headers to send during the WebSocket handshake.
    :keyword dict[str, Union[str, int]] http_proxy: HTTP proxy configuration 
    - host, port, username, password.
    :keyword list[bytes] subprotocols: Supported subprotocols.
    :keyword float timeout: Timeout value for WebSocket connection operations.
    :keyword SSLContext ssl: SSLContext object for configuring the SSL connection.
    """

    def __init__(
        self,
        url: str,
        *,
        headers: Optional[Dict[bytes, bytes]] = None,
        http_proxy: Optional[Dict[str, Union[int, str]]] = None,
        subprotocols: Optional[List[bytes]] = None,
        timeout: Optional[float] = None,
        ssl_ctx: Optional["SSLContext"] = None, # pylint: disable=unused-argument
    ) -> None:
        self._url: str = url
        self._is_open: bool = False
        self._protocol: WebSocketProtocol = WebSocketProtocol(
            url,
            headers=headers,
            subprotocols=subprotocols,
            http_proxy=http_proxy,
            ssl_ctx=ssl_ctx,
            timeout=timeout
        )
        self._headers = headers
        self._http_proxy = http_proxy
        self._subprotocols = subprotocols

    def connect(self) -> None:
        """
        Establishes a WebSocket connection with the server.
        """
        if self._is_open:
            return
        self._protocol.open_connection()
        self._is_open = True

    def send_str(self, data: str) -> None:
        """
        Sends a text message to the server. Raises a TypeError if the data is not a string.
        :param str data: The text message to send.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")
        if not isinstance(data, str):
            raise TypeError("data must be a string")

        self._protocol.send_frame(data, Opcode.TEXT)

    def send_bytes(self, data: bytes) -> None:
        """
        Sends a binary message to the server. Raises a TypeError if the data is not bytes.
        :param bytes data: The binary message to send.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")
        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")

        self._protocol.send_frame(data, Opcode.BINARY)

    def receive_str(self) -> str:
        """
        Receives the data from the server as a string.
        :returns: The received data as a string.
        :rtype: str
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        frame = self._protocol.receive_frame()

        if frame.opcode in (Opcode.TEXT, Opcode.BINARY):
            return frame.data.decode("utf-8")
        return ""

    def receive_frame(self) -> Frame:
        """
        Receives the websocket frame from the server
        :returns Frame: The frame received from the server.
        :rtype: Frame
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        return self._protocol.receive_frame()

    def receive_bytes(self) -> bytes:
        """
        Receives the data from the server as bytes.
        :returns: The data received from the server as bytes.
        :rtype: bytes
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        frame = self._protocol.receive_frame()

        if frame.opcode in (Opcode.TEXT, Opcode.BINARY):
            if isinstance(frame.data, str):
                return frame.data.encode("utf-8")
            return frame.data

        return b""

    def close(self, code: CloseReason = CloseReason.NORMAL, reason: bytes = b"") -> None:
        """
        Closes the WebSocket connection with the server.
        :param CloseReason code: The close code.
        :param bytes reason: The close reason.
        """
        if not self._is_open:
            return
        self._protocol.close(code, reason)
        self._is_open = False

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def ping(self, data: bytes = b"") -> None:
        """
        Sends a ping frame to the server.
        :param bytes data: The data to send with the ping frame.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")

        self._protocol.send_frame(data, Opcode.PING)

    def pong(self, data: bytes = b"") -> None:
        """
        Sends a pong frame to the server.
        :param bytes data: The data to send with the pong frame.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")

        self._protocol.send_frame(data, Opcode.PONG)


class AsyncWebSocket:
    """
    An async WebSocket client implementation that provides a simple interface 
    for sending and receiving data.

    :param str url: The URL of the WebSocket server. It must start with either ws:// or wss://
    and in the format ws://host:port/path or wss://host:port/path.
    :keyword dict[bytes, bytes] headers: A dictionary of HTTP headers to send
    during the WebSocket handshake.
    :keyword dict[str, Union[str, int]] http_proxy: A dictionary of HTTP proxy configuration
    - host, port, username, password.
    :keyword list[bytes] subprotocols: A list of supported subprotocols
    :keyword float timeout: The timeout value for the WebSocket connection to perform send
    and receive operations.
    :keyword SSLContext ssl: An SSLContext object that is used to configure the SSL connection.
    """

    def __init__(
        self,
        url: str,
        *,
        headers: Optional[Dict[bytes, bytes]] = None,
        http_proxy: Optional[Dict[str, Union[int, str]]] = None,
        subprotocols: Optional[List[str]] = None,
        timeout: Optional[float] = None, # pylint: disable=unused-argument
        ssl_ctx: Optional["SSLContext"] = None,
    ) -> None:
        self._url: str = url
        self._is_open: bool = False
        self._protocol: AsyncWebSocketProtocol = AsyncWebSocketProtocol(
            url,
            headers=headers,
            subprotocols=subprotocols,
            http_proxy=http_proxy,
            ssl_ctx = ssl_ctx,
        )
        self._headers = headers
        self._http_proxy = http_proxy
        self._subprotocols = subprotocols

    async def connect(self) -> None:
        """
        Establishes a WebSocket connection with the server.
        """
        if self._is_open:
            return
        await self._protocol.open_connection()
        self._is_open = True

    async def send_str(self, data: str) -> None:
        """
        Sends a text message to the server. Raises a TypeError if the data is not a string.
        :param str data: The text message to send.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")
        if not isinstance(data, str):
            raise TypeError("data must be a string")

        await self._protocol.send_frame(data, Opcode.TEXT)

    async def send_bytes(self, data: bytes) -> None:
        """
        Sends a binary message to the server. Raises a TypeError if the data is not bytes.
        :param bytes data: The binary message to send.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")
        if not isinstance(data, bytes):
            raise TypeError("data must be a string")

        await self._protocol.send_frame(data, Opcode.BINARY)

    async def receive_str(self) -> str:
        """
        Receives the data from the server as a string.
        :return: The received data as a string.
        :rtype: str
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        frame = await self._protocol.receive_frame()

        if frame.opcode in (Opcode.TEXT, Opcode.BINARY):
            return frame.data.decode("utf-8")
        return ""

    async def receive_frame(self) -> Frame:
        """
        Receives the websocket frame from the server
        :return Frame: The frame received from the server.
        :rtype: Frame
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        return await self._protocol.receive_frame()

    async def receive_bytes(self) -> bytes:
        """
        Receives the data from the server as bytes.
        :returns: The data received from the server.
        :rtype: bytes
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        frame = await self._protocol.receive_frame()

        if frame.opcode in (Opcode.TEXT, Opcode.BINARY):
            if isinstance(frame.data, str):
                return frame.data.encode("utf-8")
            return frame.data
        if frame.opcode == Opcode.BINARY:
            return frame.data

        return b""

    async def close(self, code: CloseReason = CloseReason.NORMAL, reason: bytes = b"") -> None:
        """
        Closes the WebSocket connection with the server.
        :param CloseReason code: The close code.
        :param bytes reason: The close reason.
        """
        if not self._is_open:
            return
        await self._protocol.close(code, reason)
        self._is_open = False

    async def __aenter__(self) -> "AsyncWebSocket":
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def ping(self, data: bytes = b"") -> None:
        """
        Sends a ping frame to the server.
        :param bytes data: The data to send with the ping frame.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")

        await self._protocol.send_frame(data, Opcode.PING)

    async def pong(self, data: bytes = b"") -> None:
        """
        Sends a pong frame to the server.
        :param bytes data: The data to send with the pong frame.
        """
        if not self._is_open:
            raise WebSocketConnectionError("Connection is closed")

        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")

        await self._protocol.send_frame(data, Opcode.PONG)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import asyncio
import errno
import socket
import ssl
import struct
from typing import Union
from ._url import parse_url, WebSocketURL

from ._headers import (
    build_request_headers,
    parse_response_headers,
    build_key,
    match_key,
    parse_proxy_response
)
from ._constants import EOF, ConnectionStatus
from ._frame import Frame, Opcode, CloseReason
from ._exceptions import WebSocketConnectionError, WebSocketPayloadError, WebSocketProtocolError


class WebSocketMixin:
    def validate_handshake(self, code: bytes, status: bytes, headers: dict) -> None:
        self._key: bytes
        if code != b'101':
            raise ConnectionError(f'Handshake failed: {status.decode()}')

        if b'upgrade' in headers and headers[b'upgrade'] != b'websocket':
            raise ConnectionError('Server did not upgrade to websocket')

        if b'connection' in headers and headers[b'connection'] != b'upgrade':
            raise ConnectionError('Connection was not upgraded to websocket')

        if b'sec-websocket-accept' in headers:
            if not match_key(self._key, headers[b'sec-websocket-accept']):
                raise ConnectionError('Server did not accept the key')


class WebSocketProtocol(WebSocketMixin):
    def __init__(self,
                 url: str,
                 *,
                 headers = None,
                 http_proxy = None,
                 subprotocols = None,
                 ssl_ctx=None,
                 timeout=None
        ) -> None:
        self._socket: socket.socket
        self._url: str = url
        self._key = build_key()
        self._conn_status = ConnectionStatus.CLOSED
        self._headers = headers if headers else {}
        self._http_proxy = http_proxy
        self._subprotocols = subprotocols
        self._ws_url: WebSocketURL = parse_url(self._url)
        self._timeout = timeout
        self._ssl_ctx = ssl_ctx


    def open_connection(self) -> None:

        # create a socket connection
        try:
            # TODO: handle proxy support
            if self._http_proxy:
                self._socket = socket.create_connection(
                    (self._http_proxy['host'], self._http_proxy['port']),
                    timeout=self._timeout
                )
                self._socket.sendall(f'CONNECT {self._ws_url.hostname}:{self._ws_url.port} HTTP/1.1\r\n\r\n'.encode())
                proxy_response = parse_proxy_response(self._socket.recv(1024))
                if not proxy_response[0].startswith(b'HTTP/'):
                    self._socket.close()
                    raise ConnectionError('Received invalid response from the proxy')

                if int(proxy_response[1]) < 100 or int(proxy_response[1]) > 999:
                    self._socket.close()
                    raise ConnectionError('Received invalid status code from the proxy')

                if proxy_response[1] != b'200':
                    self._socket.close()
                    raise ConnectionError('Could not connect to the proxy')

                if proxy_response[2] != b'Connection established':
                    raise ConnectionError('Could not connect to the proxy')
            else:
                self._socket = socket.create_connection(
                    (self._ws_url.hostname, self._ws_url.port),
                    timeout=self._timeout
                )
            if self._ws_url.is_secure:
                if 'context' not in self._ssl_ctx:
                    self._socket = self._wrap_socket_sni(
                        self._socket,
                        keyfile=self._ssl_ctx.get('keyfile', None),
                        certfile=self._ssl_ctx.get('certfile', None),
                        ca_certs=self._ssl_ctx.get('ca_certs', None),
                        ciphers=self._ssl_ctx.get('ciphers', None),
                        ssl_version=self._ssl_ctx.get('ssl_version', None),
                        server_hostname = self._ws_url.hostname
                    )
                else:
                    self._socket = self._ssl_ctx['context'].wrap_socket(
                        self._socket, server_hostname=self._ws_url.hostname
                    )
            # TODO: add logging
            self._socket.setblocking(True)
            self._socket.settimeout(self._timeout)

            # build the headers
            self._headers = build_request_headers(
                self._ws_url.path if not self._ws_url.query else f'{self._ws_url.path}?{self._ws_url.query}',
                self._ws_url.hostname,
                self._ws_url.port,
                self._ws_url.is_secure,
                self._key,
                subprotocols=self._subprotocols
                )
            # start the handshake
            self._socket.sendall(self._headers)
            self._conn_status = ConnectionStatus.CONNECTING

            self._receive_handshake()
        except ssl.SSLError as sslex:
            raise sslex from None
        except socket.error as e:
            raise WebSocketConnectionError('Could not send the handshake') from e

    def _receive_handshake(self) -> None:
        if not self._socket:
            raise ConnectionError('Socket is not connected')

        response_headers = []
        while True:
            # a response header ends with a double CRLF
            # we want to process a line at a time
            line = []
            data = self._socket.recv(1)
            while data != b'\n':
                line.append(data)
                data = self._socket.recv(1)
            line.append(data)

            header_line = b''.join(line)
            if header_line == b'\r\n':
                break
            response_headers.append(header_line)

        code, status, headers = parse_response_headers(b"".join(response_headers))

        self.validate_handshake(code, status, headers)

        self._conn_status = ConnectionStatus.OPEN

    def _wrap_socket_sni(
        self,
        sock,
        keyfile=None,
        certfile=None,
        cert_reqs=ssl.CERT_REQUIRED,
        ca_certs=None,
        do_handshake_on_connect=True,
        suppress_ragged_eofs=True,
        server_hostname=None,
        ciphers=None,
        ssl_version=None,
    ):
        """Socket wrap with SNI headers.

        Default `ssl.wrap_socket` method augmented with support for
        setting the server_hostname field required for SNI hostname header

        :param socket.socket sock: socket to wrap
        :param str or None keyfile: key file path
        :param str or None certfile: cert file path
        :param int cert_reqs: cert requirements
        :param str or None ca_certs: ca certs file path
        :param bool do_handshake_on_connect: do handshake on connect
        :param bool suppress_ragged_eofs: suppress ragged EOFs
        :param str or None server_hostname: server hostname
        :param str or None ciphers: ciphers to use
        :param int or None ssl_version: ssl version to use
        :return: wrapped socket
        :rtype: SSLSocket
        """
        # Setup the right SSL version; default to optimal versions across
        # ssl implementations
        if ssl_version is None:
            ssl_version = ssl.PROTOCOL_TLS_CLIENT
        purpose = ssl.Purpose.SERVER_AUTH

        opts = {
            "sock": sock,
            "do_handshake_on_connect": do_handshake_on_connect,
            "suppress_ragged_eofs": suppress_ragged_eofs,
            "server_hostname": server_hostname,
        }

        context = ssl.SSLContext(ssl_version)

        if ca_certs is not None:
            try:
                context.load_verify_locations(ca_certs)
            except FileNotFoundError as exc:
                exc.filename = {"ca_certs": ca_certs}
                raise exc from None
        elif context.verify_mode != ssl.CERT_NONE:
            # load the default system root CA certs.
            context.load_default_certs(purpose=purpose)

        if certfile is not None:
            context.load_cert_chain(certfile, keyfile)

        if ciphers is not None:
            context.set_ciphers(ciphers)

        if cert_reqs == ssl.CERT_NONE and server_hostname is None:
            context.check_hostname = False
            context.verify_mode = cert_reqs

        sock = context.wrap_socket(**opts)
        return sock

    def receive_frame(self) -> Frame:
        last_frame = None

        while True:
            if not self._socket or self._conn_status == ConnectionStatus.CLOSED:
                raise ConnectionError('Socket is not connected')

            # read the first two bytes to get the header of the frame
            header = self._read_socket(2)

            # parse the header
            fin = bool(header[0] >> 7 & 1)
            rsv1 = header[0] >> 6 & 1
            rsv2 = header[0] >> 5 & 1
            rsv3 = header[0] >> 4 & 1
            opcode = header[0] & 0b00001111
            mask = bool(header[1] >> 7 & 1)
            payload_len = header[1] & 0b01111111

            if payload_len == 126:
                payload_len = struct.unpack('!H', self._read_socket(2))[0]
            elif payload_len == 127:
                payload_len = struct.unpack('!Q', self._read_socket(8))[0]
            elif payload_len > 127:
                raise WebSocketPayloadError('Payload length is too large')

            masking_key = None # pylint: disable=unused-variable
            if mask:
                masking_key = self._read_socket(4)

            payload = self._read_socket(payload_len)

            frame = Frame(payload, opcode, fin, mask, rsv1, rsv2, rsv3)

            if frame.opcode == Opcode.CONTINUATION and not last_frame:
                if frame.fin:
                    raise WebSocketProtocolError("Fragements must not have FIN bit set")

            if last_frame and not last_frame.fin and frame.opcode in (Opcode.CONTINUATION, Opcode.TEXT, Opcode.BINARY):
                if frame.opcode != Opcode.CONTINUATION:
                    raise WebSocketProtocolError("Expected continuation frame")
                frame.data = last_frame.data + frame.data
                frame.opcode = last_frame.opcode

            frame.validate()

            if frame.opcode == Opcode.PING:
                self.send_frame(frame.data, Opcode.PONG)
                continue

            if frame.fin:
                break

            last_frame = frame

        if frame.opcode == Opcode.CLOSE:
            self.close()

        return frame

    def _read_socket(self, n: int) -> bytes:
        data = []

        to_receive = n

        while to_receive:
            try:
                chunk = self._socket.recv(min(4096,to_receive))
                if not chunk:
                    raise WebSocketConnectionError("Connection to the server was closed")
                data.append(chunk)
                to_receive -= len(chunk)
            except socket.error as exc:
                if isinstance(exc, ssl.SSLError) and "timed out" in str(exc):
                    raise WebSocketConnectionError("Connection timed out") from exc
                raise
        return b''.join(data)

    def send_frame(self, data: Union[bytes, str], opcode: int = 1) -> None:
        if not self._socket or self._conn_status != ConnectionStatus.OPEN:
            raise ConnectionError('Socket is not connected')

        if isinstance(data, str):
            data = data.encode("utf-8")

        frame = Frame(data,opcode)

        payload = frame.encode()

        while payload:
            try:
                sent = self._socket.send(payload)
            except ValueError:
                sent = 0
            except socket.error as exc:
                error_code = exc.args[0]
                if error_code not in [errno.EAGAIN, errno.EWOULDBLOCK]:
                    raise

            if not sent:
                raise WebSocketConnectionError("Could not send the frame")
            payload = payload[sent:]

    def close(self, code: int = CloseReason.NORMAL, reason: bytes = b"") -> None:
        if self._conn_status != ConnectionStatus.OPEN:
            return

        if not self._socket:
            raise WebSocketConnectionError('Websocket is not connected')

        if code not in CloseReason:
            raise ValueError("Invalid close code")

        close_frame = Frame(struct.pack('!H', code) + reason, Opcode.CLOSE)
        self.send_frame(close_frame.data, close_frame.opcode)
        self._conn_status = ConnectionStatus.CLOSING

        while True:
            try:
                frame = self.receive_frame()
                if frame.opcode == Opcode.CLOSE:
                    break
            except: # pylint: disable=bare-except
                break

        self._conn_status = ConnectionStatus.CLOSED
        # shut down the socket
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except OSError as e:
            # on macOS this can result in the following error:
            # OSError: [Errno 57] Socket is not connected
            # this can be ignored and is due to not handling SIGINT properly.
            if e.errno != errno.ENOTCONN:
                raise


class AsyncWebSocketProtocol(WebSocketMixin):
    def __init__(self, url: str, *, headers = None, http_proxy = None, subprotocols = None, ssl_ctx=None) -> None:
        self._stream_reader: asyncio.StreamReader
        self._stream_writer: asyncio.StreamWriter
        self._url = url
        self._key = build_key()
        self._conn_status = ConnectionStatus.CLOSED
        self._headers = headers
        self._http_proxy = http_proxy
        self._subprotocols = subprotocols
        self._ws_url: WebSocketURL = parse_url(self._url)
        self._ssl = ssl_ctx

    async def open_connection(self) -> None:
        try:
            # TODO: handle proxy support
            ssl_ctx = None
            if self._ws_url.is_secure:
                if not self._ssl:
                    ssl_ctx = self._wrap_socket_sni(server_hostname=self._ws_url.hostname)
                else:
                    ssl_ctx = self._ssl

            if self._http_proxy:
                self._stream_reader, self._stream_writer = await asyncio.open_connection(
                    self._http_proxy['host'],
                    self._http_proxy['port'],
                    ssl = ssl_ctx,
                    server_hostname = self._ws_url.hostname if ssl_ctx else None
                )
                self._stream_writer.write(
                    f'CONNECT {self._ws_url.hostname}:{self._ws_url.port} HTTP/1.1\r\n\r\n'.encode()
                )
                await self._stream_writer.drain()
                proxy_response = parse_proxy_response(await self._stream_reader.readuntil(EOF))
                if not proxy_response[0].startswith(b'HTTP/'):
                    self._stream_writer.close()
                    raise ConnectionError('Received invalid response from the proxy')

                if int(proxy_response[1]) < 100 or int(proxy_response[1]) > 999:
                    self._stream_writer.close()
                    raise ConnectionError('Received invalid status code from the proxy')

                if proxy_response[1] != b'200':
                    self._stream_writer.close()
                    raise ConnectionError('Could not connect to the proxy')

                if proxy_response[2] != b'Connection established':
                    self._stream_writer.close()
                    raise ConnectionError('Could not connect to the proxy')
            else:
                self._stream_reader, self._stream_writer = await asyncio.open_connection(
                    self._ws_url.hostname,
                    self._ws_url.port,
                    ssl = ssl_ctx,
                    server_hostname = self._ws_url.hostname if ssl_ctx else None
                    )

            # build the headers
            self._headers = build_request_headers(
                self._ws_url.path if not self._ws_url.query else f'{self._ws_url.path}?{self._ws_url.query}',
                self._ws_url.hostname,
                self._ws_url.port,
                self._ws_url.is_secure,
                self._key,
                subprotocols=self._subprotocols
                )
            # start the handshake
            self._stream_writer.write(self._headers)
            await self._stream_writer.drain()
            self._conn_status = ConnectionStatus.CONNECTING

            await self._receive_handshake()
        except socket.error as exc:
            raise WebSocketConnectionError('Could not send the handshake') from exc

    async def _receive_handshake(self) -> None:
        if not self._stream_reader:
            raise ConnectionError('Socket is not connected')

        response_headers = b''
        response_headers = await self._stream_reader.readuntil(EOF)
        code, status, headers = parse_response_headers(response_headers)

        self.validate_handshake(code, status, headers)
        self._conn_status = ConnectionStatus.OPEN

    def _wrap_socket_sni(
        self,
        keyfile=None,
        certfile=None,
        cert_reqs=ssl.CERT_REQUIRED,
        ca_certs=None,
        server_hostname=None,
        ciphers=None,
        ssl_version=None,
    ):
        """Socket wrap with SNI headers.

        Default `ssl.wrap_socket` method augmented with support for
        setting the server_hostname field required for SNI hostname header

        :param str or None keyfile: key file path
        :param str or None certfile: cert file path
        :param int cert_reqs: cert requirements
        :param str or None ca_certs: ca certs file path
        :param str or None server_hostname: server hostname
        :param str or None ciphers: ciphers to use
        :param int or None ssl_version: ssl version to use
        :return: wrapped socket
        :rtype: SSLSocket
        """
        # Setup the right SSL version; default to optimal versions across
        # ssl implementations
        if ssl_version is None:
            ssl_version = ssl.PROTOCOL_TLS_CLIENT
        purpose = ssl.Purpose.SERVER_AUTH

        context = ssl.SSLContext(ssl_version)

        if ca_certs is not None:
            try:
                context.load_verify_locations(ca_certs)
            except FileNotFoundError as exc:
                exc.filename = {"ca_certs": ca_certs}
                raise exc from None
        elif context.verify_mode != ssl.CERT_NONE:
            # load the default system root CA certs.
            context.load_default_certs(purpose=purpose)

        if certfile is not None:
            context.load_cert_chain(certfile, keyfile)

        if ciphers is not None:
            context.set_ciphers(ciphers)

        if cert_reqs == ssl.CERT_NONE and server_hostname is None:
            context.check_hostname = False
            context.verify_mode = cert_reqs

        return context

    async def receive_frame(self) -> Frame:
        last_frame = None

        while True:
            if not self._stream_reader or self._conn_status == ConnectionStatus.CLOSED:
                raise ConnectionError('Socket is not connected')

            # read the first two bytes to get the header of the frame
            header = await self._stream_reader.readexactly(2)
            # parse the header
            fin = bool(header[0] >> 7 & 1)
            rsv1 = header[0] >> 6 & 1
            rsv2 = header[0] >> 5 & 1
            rsv3 = header[0] >> 4 & 1
            opcode = header[0] & 0b00001111
            mask = bool(header[1] >> 7 & 1)
            payload_len = header[1] & 0b01111111
            if payload_len == 126:
                payload_len = struct.unpack('!H', await self._stream_reader.readexactly(2))[0]
            elif payload_len == 127:
                payload_len = struct.unpack('!Q', await self._stream_reader.readexactly(8))[0]
            elif payload_len > 127:
                raise WebSocketPayloadError('Payload length is too large')
            if rsv1 or rsv2 or rsv3:
                raise WebSocketProtocolError("Reserved bits are set")
            if opcode == Opcode.PING and payload_len > 125:
                raise WebSocketProtocolError("PING frame too large")

            masking_key = None # pylint: disable=unused-variable
            if mask:
                masking_key = await self._stream_reader.readexactly(4)
            payload = await self._stream_reader.readexactly(payload_len)
            frame = Frame(payload, opcode, fin, mask, rsv1, rsv2, rsv3)

            if frame.opcode == Opcode.CONTINUATION and not last_frame:
                if frame.fin:
                    raise WebSocketProtocolError("Fragements must not have FIN bit set")
            if last_frame and not last_frame.fin and frame.opcode in (Opcode.CONTINUATION, Opcode.TEXT, Opcode.BINARY):
                if frame.opcode != Opcode.CONTINUATION:
                    raise WebSocketProtocolError("Expected continuation frame")
                frame.data = last_frame.data + frame.data
                frame.opcode = last_frame.opcode

            frame.validate()

            if frame.opcode == Opcode.PING:
                await self.send_frame(frame.data, Opcode.PONG)
                continue

            if frame.fin:
                break
            last_frame = frame

        if frame.opcode == Opcode.CLOSE and self._conn_status == ConnectionStatus.OPEN:
            await self.close()

        return frame

    async def send_frame(self, data: Union[bytes, str], opcode: int = 1) -> None:
        if not self._stream_writer or self._conn_status != ConnectionStatus.OPEN:
            raise ConnectionError('Socket is not connected')

        if isinstance(data, str):
            data = data.encode("utf-8")

        frame = Frame(data,opcode)

        payload = frame.encode()

        try:
            self._stream_writer.write(payload)
            await self._stream_writer.drain()
        except AttributeError as exc:
            raise WebSocketConnectionError("Could not send the frame") from exc

    async def close(self, code: int = CloseReason.NORMAL, reason: bytes = b"") -> None:
        if self._conn_status != ConnectionStatus.OPEN:
            return

        if not self._stream_writer or not self._stream_reader:
            raise WebSocketConnectionError('Websocket is not connected')

        if code not in CloseReason:
            raise ValueError("Invalid close code")

        close_frame = Frame(struct.pack('!H', code) + reason, Opcode.CLOSE)
        await self.send_frame(close_frame.data, close_frame.opcode)
        self._conn_status = ConnectionStatus.CLOSING

        while True:
            try:
                frame = await self.receive_frame()
                if frame.opcode == Opcode.CLOSE:
                    break
            except: # pylint: disable=bare-except
                break

        self._conn_status = ConnectionStatus.CLOSED
        # shut down the socket
        self._stream_writer.close()
        await asyncio.sleep(0)
        if hasattr(self._stream_writer, 'abort'):
            self._stream_writer.abort()
        await self._stream_writer.wait_closed()
